#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# ------------------------------------------------

# std
import os
import logging
import shutil

# third party
import blinker

# emsm
from . import argparse_
from .lib import userinput


# Data
# ------------------------------------------------

__all__ = [
    "BasePlugin"
    ]

log = logging.getLogger(__file__)


# Classes
# ------------------------------------------------

class BasePlugin(object):
    """
    This is the base class for all plugins.


    If you want to know, how to implement your own plugin, you should also
    take a look at the :mod:`plugins.hellodolly` plugin.
    """

    #: Integer with the init priority of the plugin.
    #: A higher value results in a later initialisation.
    INIT_PRIORITY = 0

    #: Integer with the finish priority of the plugin.
    #: A higher value results in a later call of the finish method.
    FINISH_PRIORITY = 0

    #: The last version number of the EMSM version that worked correctly
    #: with that plugin.
    VERSION = "0.0.0"

    #: The plugin package can be downloaded from this url.
    #:
    #: .. seealso::
    #:
    #:      * :mod:`emsm.plugins.plugins` package manager
    DOWNLOAD_URL = str()

    #: This string is displayed when the ``--long-help`` argument is used.
    DESCRIPTION = str()

    #: If ``True``, the plugin has no :meth:`argparser` and can therefore
    #: not be invoked from the command line.
    HIDDEN = False

    #: Signal, that is emitted, when a plugin has been uninstalled.
    plugin_uninstalled = blinker.signal("plugin_uninstalled")

    def __init__(self, app, name):
        """
        Initialises the configuration and the storage of the plugin.

        **Override:**

            * Extend, but do not override.
        """
        log.info("initialising '{}' ...".format(name))

        self.__app = app
        self.__name = name

        # Get the argparser for this plugin and set it up.
        if type(self).HIDDEN:
            self.__argparser = None
        else:
            self.__argparser = app.argparser().plugin_parser(name)
            self.__argparser.add_argument(
                "--long-help",
                action = argparse_.LongHelpAction,
                description = type(self).DESCRIPTION
                )
        return None

    def app(self):
        """
        Returns the parent EMSM :class:`~emsm.application.Application`
        that owns this plugin.
        """
        return self.__app

    def name(self):
        """
        Returns the name of the plugin.
        """
        return self.__name

    def conf(self):
        """
        Returns a dictionary like object that contains the configuration
        of the plugin.
        """
        # Make sure the configuration section exists.
        main_conf = self.__app.conf().main()
        if not self.__name in main_conf:
            log.info("creating configuration section for '{}'"\
                     .format(self.__name))

            main_conf.add_section(self.__name)

            log.info("created configuration section for '{}'."\
                     .format(self.__name))

        return main_conf[self.__name]

    def data_dir(self, create=True):
        """
        Returns the directory that contains all data created by the plugin
        to manage its EMSM job.

        :param bool create:
            If the directory does not exist, it will be created.

        .. seealso::

            * :meth:`emsm.core.paths.Pathsystem.plugin_data_dir`
        """
        data_dir = self.__app.paths().plugin_data(self.__name)

        # Make sure the directory exists.
        if not os.path.exists(data_dir) and create:
            log.info("creating data directory for '{}'.".format(self.__name))

            os.makedirs(data_dir)

            log.info("created data directory for '{}'.".format(self.__name))

        return data_dir

    def argparser(self):
        """
        Returns the :class:`argparse.ArgumentParser` that is used by this
        plugin.

        If :attr:`HIDDEN` is ``True``, *None* is returned, since the plugin
        has no argument parser.

        .. seealso::

            * :meth:`emsm.core.argparse_.ArgumentParser.plugin_parser`
        """
        return self.__argparser

    def _uninstall(self):
        """
        This method is called by *uninstall()* and should remove all
        data or configuration generated by the plugin.

        **Subclass:**

            * You may completly override this method.
        """
        return None

    def uninstall(self):
        """
        Called when the plugin should be uninstalled. This method
        is interactive and requires the user to confirm if and which
        data should be removed.

        The BasePlugin removes:

            * the plugin module (the *.py* file in *plugins*)
            * the plugin data directory
            * the plugin configuration

        **Subclass:**

            Subclasses should override the :meth:`_uninstall` method.

        **Signals:**

            * :attr:`plugin_uninstalled`

        .. seealso::

            * :meth:`data_dir`
            * :meth:`conf`
            * :meth:`_uninstall`
        """
        log.info("uninstalling '{}' ...".format(self.__name))

        # Make sure the user really wants to uninstall the plugin.
        if not userinput.ask("Do you really want to remove '{}'?"\
                             .format(self.__name)
                             ):
            # I did not want to implement a new exception type for this case.
            # I think KeyboardInterrupt is good enough.
            log.info("cancelled uninstallation of '{}'.".format(self.__name))
            return None

        # Remove the python module that contains the plugin.
        plugin_module = self.__app.plugins().get_module(self.__name)
        if plugin_module:
            os.remove(plugin_module.__file__)

            log.info("removed '{}' module at '{}'."\
                     .format(self.__name, plugin_module.__file__)
                     )

        # Remove the plugin data directory.
        if userinput.ask("Do you want to remove the data directory?"):
            shutil.rmtree(self.data_dir(), True)

            log.info("removed '{}' plugin data directory at '{}'."\
                     .format(self.__name, self.data_dir(create=False))
                     )

        # Remove the configuration.
        if userinput.ask("Do you want to remove the configuration?"):
            self.__app.conf().main().remove_section(self.__name)

            log.info("removed '{}' configuration section."\
                     .format(self.__name)
                     )

        # Remove the subclass stuff.
        self._uninstall()

        # Emit the signal.
        BasePlugin.plugin_uninstalled.send(self)
        return None

    def run(self, args):
        """
        The *main* method of the plugin. This method is called if the plugin
        has been invoked by the command line arguments.

        :params argparse.Namespace args:
                is an argparse.Namespace instance that contains the values
                of the parsed command line arguments.

        Subclass:

            * You may override this method.

        .. seealso::

            * :meth:`argparser`
            * :meth:`emsm.core.argparse_.ArgumentParser.args`
            * :meth:`emsm.core.plugins.PluginManager.run`
        """
        return None

    def finish(self):
        """
        Called when the EMSM application is about to finish. This method can
        be used for background jobs or clean up stuff.

        Subclass:

            * You may override this method.

        .. seealso::

            * :meth:`emsm.core.plugins.PluginManager.finish`
        """
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# ------------------------------------------------

# std
import os
import sys
import logging
import importlib.machinery

# third party
import blinker

# local
from .version import VERSION
from .base_plugin import BasePlugin


# Backward compatibility
# ------------------------------------------------

if hasattr(importlib.machinery, "SourceFileLoader"):
    def _import_module(name, path):
        loader = importlib.machinery.SourceFileLoader(name, path)
        return loader.load_module()
else:
    import imp
    _import_module = imp.load_source
    del imp


# Data
# ------------------------------------------------

__all__ = [
    "PluginException",
    "PluginImplementationError",
    "PluginOutdatedError",
    "PluginManager",
    ]

log = logging.getLogger(__file__)


# Exceptions
# ------------------------------------------------

class PluginException(Exception):
    """
    Base class for all exceptions in this module.
    """
    pass


class PluginImplementationError(PluginException):
    """
    Raised if a plugin is not correct implemented.
    """

    def __init__(self, plugin, msg):
        self.plugin = plugin
        self.msg = msg
        return None

    def __str__(self):
        temp = "The plugin '{}' is not correct implemented. {}"\
               .format(self.plugin, self.msg)
        return temp


class PluginOutdatedError(PluginException):
    """
    Raised if the version of the plugin is not compatible with
    the EMSM version.

    .. seealso::

        * http://semver.org/
    """

    def __init__(self, plugin):
        self.plugin = plugin
        return None

    def __str__(self):
        temp = "The plugin '{}' is outdated.".format(self.plugin)
        return temp


# Classes
# ------------------------------------------------

class PluginManager(object):
    """
    Loads and manages all plugins.

    If you want to write a plugin and search for the docs, take a look at the
    :mod:`~emsm.plugins.hellodolly` plugin.

    .. seealso::

        * :class:`~emsm.core.base_plugin.BasePlugin`
    """

    def __init__(self, app):
        """
        """
        self._app = app

        # Maps the module name to the module object.
        self._plugin_modules = dict()

        # Maps the module name to the plugin class.
        self._plugin_types = dict()

        # Maps the module name to the plugin instance.
        self._plugins = dict()

        # Unload the plugin when it has been uninstalled.
        #
        # See also:
        #   * BasePlugin.uninstall()
        BasePlugin.plugin_uninstalled.connect(self._uninstall)
        return None

    def get_module(self, plugin_name):
        """
        Returns the Python module object that contains the plugin with the
        name *plugin_name* or ``None`` if there is no plugin with that name.
        """
        return self._plugin_modules.get(plugin_name)

    def get_plugin_type(self, plugin_name):
        """
        Returns the plugin class for the plugin with the name *plugin_name* or
        ``None``, if there is no plugin with that name.
        """
        return self._plugin_types.get(plugin_name)

    def plugin_is_available(self, plugin_name):
        """
        Returns ``True``, if the plugin with the name *plugin_name* is
        available.
        """
        return plugin_name in self._plugin_modules

    def get_plugin_names(self):
        """
        Returns the names of all loaded plugins.
        """
        return list(self._plugin_modules.keys())

    def get_plugin(self, plugin_name):
        """
        Returns the instance of the plugin with the name *plugin_name* that is
        currently loaded and used by the EMSM.
        """
        return self._plugins.get(plugin_name)

    def get_all_plugins(self):
        """
        Returns all currently loaded plugin instances.

        .. seealso::

            * :meth:`get_plugin_names`
            * :meth:`get_plugin`
        """
        return self._plugins.values()

    def _plugin_is_outdated(self, plugin):
        """
        Returns ``True`` if the *plugin* is outdated and not compatible with
        the current EMSM version.

        .. seealso::

            * :mod:`emsm.core.version`
            * http://semver.org
        """
        app_version = VERSION.split(".")
        plugin_version = plugin.VERSION.split(".")

        # The version number is invalid.
        if len(plugin_version) < 2:
            return True
        # Only a change in the major version number means a break
        # with the API.
        elif app_version[0] != plugin_version[0]:
            return True
        else:
            return False

    def import_plugin(self, path):
        """
        Loads the plugin located at *path*.

        .. note::

            The *path* is no longer added to :attr:`sys.path` (EMSM Vers. >= 3).

        :raises PluginOutdatedError:
            when the plugin is outdated.
        :raises PluginImplementationError:
            when the plugin is not correct implemented.

        .. seealso::

            * :meth:`_plugin_is_outdated`
        """
        # The module name is the name of the plugin.
        # I assume, that a modulename always ends with '.py'.
        name = os.path.basename(path)
        name = name[:-3]

        # Try to import the module.
        try:
            module = _import_module(name, path)
        except Exception as err:
            raise PluginImplementationError(name, err)

        # Check if the module contains a plugin.
        if not hasattr(module, "PLUGIN"):
            msg = "The gloabal 'PLUGIN' variable is not defined."
            raise PluginImplementationError(name, msg)

        if not hasattr(module, module.PLUGIN):
            msg = "The plugin module '{}' does not contain the declared "\
                  "plugin class '{}'".format(name, module.PLUGIN)
            raise PluginImplementationError(name, msg)

        # Get the plugin class.
        plugin_type = getattr(module, module.PLUGIN)

        # The plugin has to be a subclass of BasePlugin.
        if not issubclass(plugin_type, BasePlugin):
            msg = "The plugin '{}' is not a subclass of BasePlugin."\
                  .format(module.PLUGIN)
            raise PluginImplementationError(name, msg)

        # Check if the plugin is tested and compatible with the current
        # EMSM version.
        if self._plugin_is_outdated(plugin_type):
            raise PluginOutdatedError(name)

        # Save the plugin module and class.
        # A plugin instance is created later, when it is first needed.
        self._plugin_modules[name] = module
        self._plugin_types[name] = plugin_type
        return None

    def import_from_directory(self, directory):
        """
        Imports all Python modules in the :file:`directory`.

        Files that do not contain a valid EMSM plugin, are ignored. You can
        check the log files to see which plugins have been ignored.

        .. seealso::

            * :meth:`import_plugin`
        """
        def file_is_plugin(path):
            """
            Returns ``True`` if the path probably points to a plugin module.
            """
            filename = os.path.basename(path)
            if os.path.isdir(path):
                return False
            elif filename.startswith("_"):
                return False
            elif not filename.endswith(".py"):
                return False
            elif filename.count(".") != 1:
                return False
            return True

        log.info("loading plugins from '{}' ...".format(directory))

        for filename in os.listdir(directory):
            path = os.path.join(directory, filename)
            if not file_is_plugin(path):
                continue

            try:
                self.import_plugin(path)
            except PluginImplementationError as err:
                log.warning(err)
            except PluginOutdatedError as err:
                log.warning(err)
            else:
                log.info("loaded plugin from '{}'.".format(path))
        return None

    def remove_plugin(self, plugin_name, call_finish=False):
        """
        Unloads the plugin with the name *plugin_name*.

        :param str plugin_name:
            The name of the plugin that should be unloaded.
        :param bool call_finish:
            If true, the :meth:`~emsm.core.base_plugin.BasePlugin.finish`
            method of the plugin is called, before it is unloaded.
        """
        log.info("unloading plugin '{}' ...".format(plugin_name))

        plugin = self._plugins.get(plugin_name, None)

        # Break if there is not plugin with that name.
        if plugin is None:
            return None

        if call_finish:
            plugin.finish()

        # Remove the plugin.
        del self._plugins[plugin_name]
        del self._plugin_types[plugin_name]
        del self._plugin_modules[plugin_name]

        # The plugin has been removed.
        log.info("the plugin '{}' has been unloaded.".format(plugin_name))
        return None

    def _uninstall(self, plugin):
        """
        Called, when the plugin has been uninstalled.

        .. seealso::

            * :attr:`emsm.core.base_plugin.BasePlugin.plugin_uninstalled`
        """
        # Break if we do not own this plugin.
        if not plugin in self._plugins.values():
            return None

        # Unload the plugin.
        self.remove_plugin(plugin_name=plugin.name(), call_finish=False)
        return None

    def setup(self):
        """
        Imports all plugins from the application's plugin directory.

        .. seealso::

            * :meth:`emsm.core.paths.Pathsystem.plugins`
            * :meth:`emsm.core.paths.Pathsystem.emsm_plugins`
        """
        paths = self._app.paths()
        self.import_from_directory(paths.emsm_plugins())
        self.import_from_directory(paths.plugins())
        return None

    def init_plugins(self):
        """
        Creates a plugin instance for each loaded plugin class.

        When you call this method multiple times, only plugins that have
        not been initialised already, will be initialised.
        """
        log.info("initialising plugins ...")

        # Initialise the plugins corresponding to their *INIT_PRIORITY*
        init_queue = self._plugin_types.items()
        init_queue = sorted(init_queue, key=lambda e: e[1].INIT_PRIORITY)

        for name, plugin_type in init_queue:

            # The plugin has already been initialised.
            if name in self._plugins:
                continue

            # Create a new plugin instance and save it.
            plugin = plugin_type(self._app, name)
            self._plugins[name] = plugin

        log.info("initialised plugins.")
        return None

    def run(self):
        """
        Calls :meth:`~emsm.core.base_plugin.BasePlugin.run` of the plugin that
        has been selected by the command line arguments.

        .. seealso::

            * :meth:`emsm.core.argparse_.ArgumentParser.args`
        """
        # Get the name of the selected plugin.
        args = self._app.argparser().args()
        plugin_name = args.plugin

        # Break if no plugin has been selected.
        if plugin_name is None:
            log.info("no plugin for run selected.")
            return None

        # Execute the plugin.
        log.info("running plugin '{}' ...".format(plugin_name))
        plugin = self._plugins[plugin_name]
        plugin.run(args)
        return None

    def finish(self):
        """
        Calls :meth:`~emsm.core.base_plugin.BasePlugin.finish` for each loaded
        plugin.
        """
        log.info("finish plugins ...")

        finish_queue = self._plugins.values()
        finish_queue = sorted(finish_queue, key=lambda p: p.FINISH_PRIORITY)

        for plugin in finish_queue:
            plugin.finish()
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# ------------------------------------------------

#std
import time
import shutil
import os
import sys
import subprocess
import shlex
import signal
import collections
import re
import random
import socket
import logging
import io

# third party
import blinker


# Backward compatibility
# ------------------------------------------------

if not hasattr(shlex, "quote"):
    # From the Python3.3 library
    import re
    _find_unsafe = re.compile(r'[^\w@%+=:,./-]', re.ASCII).search
    def _shlex_quote(s):
        """Return a shell-escaped version of the string *s*."""
        if not s:
            return "''"
        if _find_unsafe(s) is None:
            return s

        # use single quotes, and put single quotes into double quotes
        # the string $'b is then quoted as '$'"'"'b'
        return "'" + s.replace("'", "'\"'\"'") + "'"
    shlex.quote = _shlex_quote
    del re

if not hasattr(shlex, "which"):
    shlex.which = lambda s: s

if not hasattr(subprocess, "DEVNULL"):
    subprocess.DEVNULL = os.open(os.devnull, os.O_RDWR)

try:
    FileNotFoundError
    FileExistsError
except NameError:
    FileNotFoundError = OSError
    FileExistsError = OSError


# Data
# ------------------------------------------------

__all__ = [
    "WorldError",
    "WorldStatusError",
    "WorldIsOnlineError",
    "WorldIsOfflineError",
    "WorldStartFailed",
    "WorldStopFailed",
    "WorldCommandTimeout",
    "WorldWrapper",
    "WorldManager"
    ]

log = logging.getLogger(__file__)

_SCREEN = shlex.which("screen")


# Exceptions
# ------------------------------------------------

class WorldError(Exception):
    """
    Base class for all other exceptions in this module.
    """
    pass


class WorldStatusError(WorldError):
    """
    Raised, if a task can not be done because of the
    current status of the world (online or not online).
    """

    def __init__(self, world, is_online):
        self.world = world
        self.is_online = is_online
        return None

    def __str__(self):
        if self.is_online:
            temp = "The world '{}' is online!"
        else:
            temp = "The world '{}' is offline!"
        temp = temp.format(self.world.name())
        return temp


class WorldIsOnlineError(WorldStatusError):
    """
    Raised if a world is online but should be offline.
    """

    def __init__(self, world):
        WorldStatusError.__init__(self, world, True)
        return None


class WorldIsOfflineError(WorldStatusError):
    """
    Raised if a world is offline but should be online.
    """

    def __init__(self, world):
        WorldStatusError.__init__(self, world, False)
        return None


class WorldStartFailed(WorldError):
    """
    Raised if the world start failed.
    """

    def __init__(self, world):
        self.world = world
        return None

    def __str__(self):
        temp = "The start of the world '{}' failed!"\
               .format(self.world.name())
        return temp


class WorldStopFailed(WorldError):
    """
    Raised if the world stop failed.
    """

    def __init__(self, world):
        self.world = world
        return None

    def __str__(self):
        temp = "The stop of the world '{}' failed!"\
               .format(self.world.name())
        return temp


class WorldCommandTimeout(WorldError):
    """
    Raised, when the server did not react in x seconds.
    """

    def __init__(self, world = ""):
        self.world = world
        return None

    def __str__(self):
        temp = "The world '{}' did not react!"\
               .format(self.world.name())
        return temp


# Classes
# ------------------------------------------------

class WorldWrapper(object):
    """
    Provides methods to handle a minecraft world like
    :meth:`start`, :meth:`stop`, :meth:`restart`, ...

    The WorldWrapper is initialised using the configuration options in the
    section with the name *name* in the :file:`server.conf` configuration
    file.
    """

    # Screen prefix for the minecraft-server sessions.
    # DO NOT CHANGE THIS VALUE IF A WORLD IS ONLINE!
    _SCREEN_PREFIX = "minecraft_"

    #: Signal, that is emitted when a world has been uninstalled.
    world_uninstalled = blinker.signal("world_uninstalled")

    #: Signal, that is emitted when a world is about to start.
    world_about_to_start = blinker.signal("world_about_to_start")

    #: Signal, that is emitted when a world has been started.
    world_started = blinker.signal("world_started")

    #: Signal, that is emitted when a world could not be started.
    world_start_failed = blinker.signal("world_start_failed")

    #: Signal, that is emitted when a world is about to be stopped.
    world_about_to_stop = blinker.signal("world_about_to_stop")

    #: Signal, that is emitted when a world has been stopped.
    world_stopped = blinker.signal("world_stopped")

    #: Signal, that is emitted when a world could not be stopped.
    world_stop_failed = blinker.signal("world_stop_failed")


    def __init__(self, app, name):
        """
        """
        log.info("initialising world '{}' ...".format(name))

        self._app = app

        # The name of the world.
        self._name = name

        # Get the configuration.
        self._conf = app.conf().worlds()[name]
        self._check_conf()

        # The ServerWrapper for the server that powers this world.
        self._server = app.server().get(self._conf["server"])
        if not self._server.is_installed():
            self._server.install()

        # The directory that contains the world data.
        self._directory = app.paths().world(name)
        return None

    def _check_conf(self):
        """
        Checks if the configuration contains only valid values and
        types (float, int, ...).

        :raises ValueError:
            If a configration option has an invalid value.
        :raises TypeError:
            If a configuration option has an invalid type.
        """
        # stop timeout
        if not self._conf["stop_timeout"].isdecimal():
            raise TypeError("{} - conf:stop_timeout is not a positive integer"\
                            .format(self._name))

        # stop delay
        if not self._conf["stop_delay"].isdecimal():
            raise TypeError("{} - conf:stop_delay is not a positive integer"\
                            .format(self._name))

        # server
        if not self._conf["server"] in self._app.server().get_names():
            raise ValueError("{} - conf:server does not exist"\
                             .format(self._name))
        return None

    def worldpath_to_ospath(self, rel_path):
        """
        Converts *rel_path*, that is relative to the root directory of the
        minecraft world, into the absolute path of the operating system.

        **Example:**

        .. code-block:: python

            >>> # I assume the EMSM root is: "/home/minecraft"
            >>> foo.name()
            "foo"
            >>> foo.worldpath_to_ospath("server.properties")
            "/opt/minecraft/worlds/foo/server.properties"

        .. seealso::

            * :meth:`directory`
        """
        return os.path.join(self._directory, rel_path)

    def conf(self):
        """
        The configuration section of this world.

        .. seealso::

            * :class:`~emsm.core.conf.WorldsConfiguration`
        """
        return self._conf

    def server(self):
        """
        The :class:`~emsm.core.server.ServerWrapper` for the server that runs
        this world.
        """
        return self._server

    def set_server(self, server):
        """
        Changes the server that runs this world. The world has to be offline.

        :param emsm.core.server.ServerWrapper server:
            The new server

        :raises WorldIsOnlineError*
            if the world is online.
        """
        if self.is_online():
            raise WorldIsOnlineError(self)

        # Break, if we have nothing to do.
        if server is self._server:
            return None

        self._server = server
        self._conf["server"] = server.name()

        log.info("assigned '{}' server to the world '{}'."\
                 .format(server.name(), self._name)
                 )
        return None

    def name(self):
        """
        The name of the world.

        This is the name of the configuration section in :file:`worlds.conf` and
        the folder name in the :file:`worlds` directory.
        """
        return self._name

    def screen_name(self):
        """
        Returns the name of the screen sessions that run the server of this
        world.
        """
        return WorldWrapper._SCREEN_PREFIX + self._name

    def directory(self):
        """
        Returns the directory that contains all world data generated by the
        minecraft server.

        If the world is run by the *mojang* minecraft server, this directory
        contains the :file:`server.properties`, :file:`whitelist.json`, ...
        files.
        """
        return self._directory

    def address(self):
        """
        Returns the binding (ip, port) of the world. If those values can not
        be retrieved, ``(None, None)`` is returned.
        """
        return self._server.world_address(self)

    def latest_log(self):
        """
        Returns the log of the world since the last start. If the
        logfile does not exist, an empty string will be returned.
        """
        # Matches all lines in the log, that signalize the start of
        # a server.
        re_start_line = self._server.log_start_re()
        log_path = os.path.abspath(
            os.path.join(self._directory, self._server.log_path())
            )

        try:
            last_log = io.StringIO()
            with open(log_path) as log:
                for line in log:
                    if re.match(re_start_line, line):
                        last_log = io.StringIO()
                    last_log.write(line)
            last_log = last_log.getvalue()
        except (FileNotFoundError, IOError) as err:
            last_log = str()
        return last_log

    def pids(self):
        """
        Returns a list with the pids of the screen sessions with the name
        :meth:`screen_name`.
        """
        # Get sessions
        # XXX: screen -ls seems to exit always with the exit code 1.
        #   so it's convenient to use gestatusoutput.
        status, output = subprocess.getstatusoutput("screen -ls")

        # Example output (without the '>' char):
        #
        # > foo@bar:~$ screen -ls
        # > There is a screen on:
        # >        20405.minecraft_barz    (07/08/13 14:42:15)     (Detached)
        # > 1 Socket in /var/run/screen/S-foo.

        # Filter the PIDs
        re_pid = re.compile(
            "^\s*(\d+?).{screen_name}".format(screen_name=self.screen_name()),
            re.MULTILINE
            )
        pids = re.findall(re_pid, output)
        pids = [int(pid) for pid in pids]
        return pids

    def is_online(self):
        """
        Returns ``True`` if the world is currently running.
        """
        return bool(self.pids())

    def is_offline(self):
        """
        Returns ``True`` if the world is currently **not** running.
        """
        return not self.is_online()

    def send_command(self, server_cmd):
        """
        Sends the given command to all screen sessions with the world's screen
        name.

        :raises WorldIsOfflineError:
            if the world is offline.

        .. warning::

            There is no guarantee, that the server reacted to the command.
        """
        pids = self.pids()

        # Break if the world is offline.
        if not pids:
            raise WorldIsOfflineError(self)

        # Translate the server command for *cross-server* support.
        server_cmd = self._server.translate_command(server_cmd)

        # Quote the command.
        # The '\n' simulates pressing the ENTER key in a screen session.
        server_cmd += "\n\n"
        server_cmd = shlex.quote(server_cmd)

        # Send the command to the server.
        for pid in pids:
            sys_cmd = "screen -S {0}.{1} -p 0 -X stuff {2}"\
                      .format(pid, shlex.quote(self.screen_name()), server_cmd)
            sys_cmd = shlex.split(sys_cmd)
            subprocess.call(sys_cmd)
        return None

    def send_command_get_output(self, server_cmd, timeout=10,
                                poll_intervall=0.2):
        """
        Like :meth:`send_commmand` but checks every *poll_intervall*
        seconds, if content has been added to the logfile and returns the
        change. If no change could be detected after *timeout* seconds,
        an error will be raised.

        :raises WorldIsOfflineError:
            if the world is offline.
        :raises WorldCommandTimeout:
            if the world did not react within *timeout* seconds.
        """
        log_path = os.path.abspath(
            os.path.join(self._directory, self._server.log_path())
            )

        # Save the current size of the logfile to detect changes.
        try:
            with open(log_path) as log:
                log.seek(0, 2)
                offset = log.tell()
        except (FileNotFoundError, IOError):
            offset = 0

        # Send the command.
        self.send_command(server_cmd)

        # Parse the logfile for a change.
        start_time = time.time()
        output = str()
        while (not output) and time.time() - start_time < timeout:
            time.sleep(poll_intervall)

            try:
                with open(log_path) as log:
                    log.seek(offset, 0)
                    output = log.read()
            except (FileNotFoundError, IOError):
                break

        if not output:
            raise WorldCommandTimeout(self)
        return output

    def open_console(self):
        """
        Opens **all** screen sessions whichs pid is in :meth:`pids`.

        :raises WorldIsOfflineError:
            if the world is offline.
        """
        pids = self.pids()

        # Break if the world is offline.
        if not pids:
            raise WorldIsOfflineError(self)

        # Open all world screen sessions (one for each found pid).
        for pid in pids:
            sys_cmd = "screen -x {pid}".format(pid=pid)

            try:
                subprocess.check_call(shlex.split(sys_cmd))
            except subprocess.CalledProcessError as error:
                # It's probably not the terminal of the user,
                # so try this one.
                sys_cmd = "script -c {} /dev/null"\
                          .format(shlex.quote(sys_cmd))
                subprocess.check_call(
                    shlex.split(sys_cmd),
                    stdin = sys.stdin,
                    stdout = sys.stdout,
                    stderr = sys.stderr
                    )
        return None


    def is_installed(self):
        """
        Returns ``True`` if the :meth:`directory` of the world exists,
        otherwise ``False``.

        .. seealso::

            * :meth:`directory`
            * :meth:`install`
            * :meth:`uninstall`
        """
        return os.path.exists(self._directory) \
               and os.path.isdir(self._directory)

    def install(self):
        """
        Creates the directory of the world.

        .. seealso::

            * meth:`directory`
        """
        try:
            os.makedirs(self._directory, exist_ok=True)
        # XXX: Fixes an error with sudo / su
        except FileExistsError:
            pass
        return None

    def uninstall(self):
        """
        Stops the world and removes the world directory.

        .. seealso::

            * :meth:`kill_processes`
            * :meth:`directory`
        """
        self.kill_processes()

        # Try 5 times to remove the directory. This is necessairy, if the
        # world was online and fixes a problem with *server.log.lck*.
        for i in range(5):
            try:
                shutil.rmtree(self._directory)
            except FileExistsError:
                time.sleep(0.5)
            else:
                break

        # Remove the configuration.
        self._app.conf().worlds().remove_section(self._name)

        # Emit the corresponing signal to this event.
        WorldWrapper.world_uninstalled.send(self)
        return None


    def start(self, wait_check_time=0.1):
        """
        Starts the world if the world is offline. If the world is already
        online, nothing happens.

        **Signals:**

            * :attr:`world_about_to_start`
            * :attr:`world_started`
            * :attr:`world_start_failed`

        :param float wait_check_time:
            Time waited, before checking if the server actually started.

        :raises WorldStartFailed:
            if the world could not be started.
        """
        global _SCREEN

        # Break if the world is already online.
        if self.is_online():
            return None

        WorldWrapper.world_about_to_start.send(self)

        # We need to change the current working directory to the world's
        # directory so that the server starts in the correct environment.
        old_wd = os.getcwd()
        os.chdir(self.directory())
        try:
            if os.path.samefile(os.getcwd(), self.directory()):
                sys_cmd = "{screen} -dmS {screen_name} {start_cmd}".format(
                    screen = _SCREEN,
                    screen_name = shlex.quote(self.screen_name()),
                    start_cmd = self._server.start_cmd()
                    )

                # Check if a screenrc file should be used.
                #
                # Note: Put the screenrc path into another configuration file?
                screenrc_path = self._app.conf().main()["emsm"]["screenrc"]
                if screenrc_path:
                    sys_cmd += " -c {}".format(shlex.quote(screenrc_path))

                # Fire off the start command.
                sys_cmd = shlex.split(sys_cmd)
                subprocess.call(sys_cmd)
        finally:
            # We may have not the rights to change back.
            try:
                os.chdir(old_wd)
            except OSError:
                pass

        # Check if the world is really online.
        time.sleep(wait_check_time)
        if not self.is_online():
            WorldWrapper.world_start_failed.send(self)
            raise WorldStartFailed(self)

        WorldWrapper.world_started.send(self)
        return None


    def kill_processes(self):
        """
        Kills all processes with a pid in :meth:`pids`.

        **Signals:**

            * :attr:`world_about_to_stop`
            * :attr:`world_stopped`
            * :attr:`world_stop_failed`

        :raises WorldStopFailed: if the process could not be killed.

        .. seealso::

            * :meth:`pids`
        """
        pids = self.pids()

        # Break if the world is already offline.
        if not pids:
            return None

        # Kill all processes.
        WorldWrapper.world_about_to_stop.send(self)
        for pid in pids:
            os.kill(pid, signal.SIGTERM)

        # Check if the world is now offline.
        if self.is_online():
            WorldWrapper.world_stop_failed.send(self)
            raise WorldStopFailed(self)

        WorldWrapper.world_stopped.send(self)
        return None


    def stop(self, force_stop=False, message=None, delay=None,
             timeout=None):
        """
        Stops the server.

        :param str message:
            Send to the world before the :meth:`stop` command is executed.
        :param float delay:
            Time in seconds that is waited between seding the *message*
            and executing the :meth`stop` command.
        :param float timeout:
            Maximum time in seconds waited for the server stop after
            executing the :meth:`stop` command.
        :param bool force_stop:
            If true and the server could not be stopped,
            :meth:`kill_processes` is called.

        **Signals:**

            * :attr:`world_about_to_stop`
            * :attr:`world_stopped`
            * :attr:`world_stop_failed`

        :raises WorldStopFailed:
            if the world could not be stopped.

        .. seealso::

            * :meth:`kill_processes`
            * :meth:`is_offline`
        """
        # Break if the world is already offline.
        if self.is_offline():
            return None

        WorldWrapper.world_about_to_stop.send(self)

        # Get the default parameter values from the configuration.
        if message is None:
            message = self._conf["stop_message"]
        if delay is None:
            delay = int(self._conf["stop_delay"])
        if timeout is None:
            timeout = int(self._conf["stop_timeout"])

        # Send the stop_message.
        for line in message.split("\n"):
            line = line.strip()
            self.send_command("say {}".format(line))

        # Save the world and wait delay seconds to make sure the
        # world is saved and the stop_message can be read.
        self.send_command("save-all")
        time.sleep(delay)

        # Stop the world.
        self.send_command("stop")
        start_time = time.time()
        while self.is_online() and time.time() - start_time < timeout:
            time.sleep(0.25)

        # Force the stop if necessairy.
        if force_stop:
            self.kill_processes()

        # Check if the world is offline.
        if self.is_online():
            WorldWrapper.world_stop_failed.send(self)
            raise WorldStopFailed(self)

        WorldWrapper.world_stopped.send(self)
        return None

    def restart(self, force_restart=False, stop_args=None):
        """
        Restarts the server.

        :param bool force_restart:
            Forces the stop of the server by calling :meth:`kill_processes`` if
            necessairy.
        :param dict stop_args:
            If provided, these values are passed to :meth:`stop`.

        **Signals:**

            * :attr:`world_about_to_stop`
            * :attr:`world_stopped`
            * :attr:`world_stop_failed`
            * :attr:`world_about_to_start`
            * :attr:`world_started`
            * :attr:`world_start_failed`

        :raises WorldStopFailed:
            if the world could not be stopped.
        :raises WorldStartFailed:
            if the world could not be restarted.

        .. seealso::

            * :meth:`stop`
            * :meth:`start`
        """
        if stop_args is None:
            stop_args = dict()

        self.stop(force_stop=force_restart, **stop_args)
        self.start()
        return None


class WorldManager(object):
    """
    Works as a container for the :class:`WorldWrapper` instances.
    """

    def __init__(self, app):
        self._app = app

        # Maps the name of the world to the world wrapper
        # world.name() => world
        self._worlds = dict()

        WorldWrapper.world_uninstalled.connect(self._remove)
        return None

    def load_worlds(self):
        """
        Loads all worlds declared in the :file:`worlds.conf` configuration
        file.

        .. seealso::

            * :class:`~emsm.core.conf.WorldsConfiguration`
        """
        conf = self._app.conf().worlds()
        for section in conf.sections():
            world = WorldWrapper(self._app, section)
            self._worlds[world.name()] = world

            # Make sure the folder exists.
            if not world.is_installed():
                world.install()
        return None

    # container
    # --------------------------------------------

    def _remove(self, world):
        """
        Removes the :class:`WorldWrapper` *world* from the internal map.
        """
        if world.name() in self._worlds:
            del self._worlds[world.name()]
        return None

    def get(self, worldname):
        """
        Returns the :class:`WorldWrapper` for the world with the name
        *worldname* or ``None`` if there is no world with that name.
        """
        return self._worlds.get(worldname)

    def get_all(self):
        """
        Returns a list with all loaded worlds.
        """
        return list(self._worlds.values())

    def get_by_pred(self, pred=None):
        """
        Filters the worlds using the predicate *pred*.

        **Example:**

        .. code-block:: python

            >>> # All running worlds
            >>> wm.get_by_pred(lambda w: w.is_online())
            ...

        .. seealso::

            * :meth:`get_all`
        """
        return list(filter(pred, self._worlds.values()))

    def get_selected(self):
        """
        Returns all worlds that have been selected per command line argument.

        .. seealso::

            * :meth:`emsm.core.argparse_.ArgumentParser.args`
        """
        args = self._app.argparser().args()

        selected_worlds = args.worlds
        all_worlds = args.all_worlds

        if all_worlds:
            return list(self._worlds.values())
        else:
            return [self._worlds[world] for world in selected_worlds]

    def get_names(self):
        """
        Returns a list with the names of all worlds.
        """
        return list(self._worlds.keys())



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
About
-----

This plugins works as a tutorial. It's inspired by the wordpress plugin
`Hello Dolly <https://wordpress.org/plugins/hello-dolly/>`_.

Code and Download
-----------------

You can find the latest version of *hello_dolly* on the EMSM GitHub
`GitHub repository <https://github.com/benediktschmitt/emsm/blob/master/plugins/plugins.py>`_.

..
    By the way, this is a comment block in reST.

    The next line is a little hack. Unfortunetly, sphinx does not find
    "hellodolly.py" when this docstring is included with autodoc. So this does
    not work:

    .. literalinclude:: hellodolly.py

    The next line is actually a small hack. When the documentation is
    built, this module is included from
    ``EMSM_ROOT/docs/source/plugins/``, but the module is in
    ``EMSM_ROOT/plugins/```.

.. literalinclude:: ../../../emsm/plugins/hellodolly.py

Installation
------------

We want to distribute our plugin because we think it brings so much joy
to all players. So let's create a small package.

This is usually done with the :mod:`plugins.plugins` plugin:

.. code-block:: bash

    $ foo@bar: ls
    hellodolly.py ...
    $ foo@bar: plugin.py --source hellodolly.py
    $ foo@bar: ls
    hellodolly.py hellodolly.tar.bz2 ...

The compressed package archive should now be in your working directory.

Usage
-----

.. code-block:: bash

    $ foo@bar: # Will print only one row:
    $ foo@bar: minecraft -W hellodolly

    $ foo@bar: # Prints 5 rows or less, if the configuration value is smaller:
    $ foo@bar: minecraft -W hellodolly --rows 5

Documentation
-------------

Acutally, EMSM uses sphinx *autodoc* feature to create the documentation for
the plugins. So what, you see here is the docstring of the ``hellodolly.py``
module.
"""


# Modules
# ------------------------------------------------

# std
import os
import random

# third party
import termcolor

# local
import emsm
from emsm.core.base_plugin import BasePlugin


# Data
# ------------------------------------------------

# This variable helps the EMSM to find the actual plugin class in this module.
PLUGIN = "HelloDolly"

# These are the well-known hello dolly lyrics.
LYRICS = """Hello, Dolly
Well, hello, Dolly
It's so nice to have you back where you belong
You're lookin' swell, Dolly
I can tell, Dolly
You're still glowin', you're still crowin'
You're still goin' strong
We feel the room swayin'
While the band's playin'
One of your old favourite songs from way back when
So, take her wrap, fellas
Find her an empty lap, fellas
Dolly'll never go away again
Hello, Dolly
Well, hello, Dolly
It's so nice to have you back where you belong
You're lookin' swell, Dolly
I can tell, Dolly
You're still glowin', you're still crowin'
You're still goin' strong
We feel the room swayin'
While the band's playin'
One of your old favourite songs from way back when
Golly, gee, fellas
Find her a vacant knee, fellas
Dolly'll never go away
Dolly'll never go away
Dolly'll never go away again"""


# Classes
# ------------------------------------------------

class HelloDolly(BasePlugin):

    # We don't need to wait for other plugins, so we don't care
    # about the init priority. If you want that your plugin is initialised
    # earlier than others, make this value smaller.
    INIT_PRIORITY = 0

    # Also, we don't care about if the finish method of our plugin is called
    # early or late. The *finish* method of plugins with a smaller
    # *FINISH_PRIORITY* is called earlier.
    FINISH_PRIORITY = 0

    # At the moment, there is no direct url to the latest version of this
    # plugin.
    # In the future, the plugin manager could use this url to detect new
    # versions of your plugin and will download them automatically.
    DOWNLOAD_URL = None

    # The last compatible version of the EMSM.
    VERSION = "4.0.0-beta"

    # The EMSM automatically uses the DESCRIPTION variable to set up the
    # *--long-help* argument parser argument.
    #
    # We ususally use here the module's docstring. Note, that ``__doc__``
    # does not interfere with the HelloDolly docstring ``HelloDolly.__doc__``
    # since the HelloDolly class has no docstring.
    DESCRIPTION = __doc__

    def __init__(self, application, name):
        """
        """
        # We need to init the BasePlugin. This is necessairy, so that we can
        # safely access:
        #
        #   * self.conf()
        #   * self.argparser()
        #   * ...
        BasePlugin.__init__(self, application, name)

        # The configuration and argument parser are set up in own methods
        # for readability.
        self._setup_conf()
        self._setup_argparser()
        return None

    def _setup_conf(self):
        """
        Sets the configuration up.
        """
        # Get the configuration dictionary for this plugin.
        conf = self.conf()

        # This is an example of the hellodolly configuration section in the
        # main.conf configuration file:
        #
        # [hellodolly]
        # max_rows = 5
        #

        self._max_rows = conf.getint("max_rows", 5)
        conf["max_rows"] = str(self._max_rows)
        return None

    def _setup_argparser(self):
        """
        Sets the argument parser up.
        """
        # Get the plugin's argument parser.
        parser = self.argparser()

        parser.description = (
            "Demonstrates the implementation of a plugin. Inspired by the "
            "wordpress plugin \"Hello, Dolly\"."
            )
        parser.epilog = "https://github.com/benediktschmitt/emsm"

        # Note, that we prefix the *dest* value, since all arguments share
        # the same namespace.
        parser.add_argument(
            "--rows", "-r",
            action = "store",
            dest = "hellodolly_rows",
            type = int,
            default = 1,
            metavar = "ROWS",
            help = "The number of lines that will be printed."
            )
        return None

    def _uninstall(self):
        """
        If you created data not stored in ``data_dir()`` or used also the
        *worlds.conf* or *server.conf* configuration files, you should ask the
        user here, if he wants to remove these files and settings too.

        Note the difference between ``_uninstall()`` and ``uninstall()``.
        """
        # Your uninstallation stuff here
        # ...
        return None

    def run(self, args):
        """
        Writes lines of our lyrics into the chats of the selected worlds.

        Parameters:
            * args
                Is a namespace that contains the parsed arguements.
        """
        # Get the number of lines we want to print and make sure, that
        # the number is not greater then the max_rows configuration value.
        rows = args.hellodolly_rows
        if rows > self._max_rows:
            rows = self._max_rows
        if rows < 0:
            rows = 0

        # Run hellodolly for each world, which has been selected with
        # *-w* or *-W* per command line.
        # We sort the worlds by their names, to process them in alphabetical
        # order.
        worlds = self.app().worlds().get_selected()
        worlds.sort(key = lambda w: w.name())

        for world in worlds:
            self.be_poetic(world, rows)
        return None

    def get_lyrics(self, num_rows):
        """
        Returns rows of the hello dolly lyrics.

        Parameters:
            * num_rows
                The number of rows, that should be extracted from the
                lyrics.
        """
        global LYRICS
        lyrics = LYRICS

        # Get *num_rows* lines of the lyrics.
        lyrics = lyrics.split("\n")
        if num_rows > len(lyrics):
            return lyrics
        else:
            a = random.randint(0, len(lyrics) - num_rows)
            lyrics = lyrics[a:a+num_rows]
        return lyrics

    def be_poetic(self, world, num_rows):
        """
        Writes the *lyrics* to the chat of all running, selected worlds.
        """
        lyrics = self.get_lyrics(num_rows)

        # We follow the inofficial EMSM style guide and print the
        # world name in cyan.
        print(termcolor.colored("{}:".format(world.name()), "cyan"))
        if world.is_offline():
            print("\t", termcolor.colored("error:", "red"), "world is offline")
        else:
            for row in lyrics:
                world.send_command("say {}".format(row))
            print("\t", "world has been visited")
        return None

    def finish(self):
        """
        This method is always called, when the EMSM is about to finish.
        It should be used for clean up or background stuff.
        """
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
About
-----

Monitors selected worlds (*--world*, *-w*, *-W*) and reacts on issues.

Download
--------

You can find the latest version of this plugin in the EMSM
`GitHub repository <https://github.com/benediktschmitt/emsm>`_.

Configuration
-------------

Since EMSM version 3.2.2-beta, this plugin requires no more configuration.
The command line arguments allow you to adjust the guard for each world.

Arguments
---------

When invoked all worlds selected with the global EMSM commands *-W* or *-w*
are checked.

.. option:: --error-action {none, stop, restart}

    Defines how the guard handles a world in trouble.

.. note::

    Per default, all tests will be performed. If you don't want to run all
    tests, you can pass the tests, which should be performed as command
    line arguments.

.. option:: --test-status

    Check if the world is online.

.. option:: --test-log

    Check if the logs contain an error.

.. option:: --test-port

    Check if the world is reachable.

.. option:: --output-format {console, text}

    Defines the output format.

    *text* is suitable for sending the guard output via email.

.. option:: --output-only-new-warnings

    If an error with the world has been detected in the previous run, the
    warning for this world will be suppressed.

Cron
----

This plugin is made for cron (therefore it does not print much):

.. code-block:: text

    # m h dom mon dow user command
    # Runs the guard every 5 minutes for all worlds
    */5 * *   *   *   root minecraft -W guard --output-only-new-warnings --output-format text

    # Runs the guard every 5 minutes for the world *foo*.
    */5 * *   *   *   root minecraft -w foo guard --output-only-new-warnings --output-format text

Changelog
---------

3.0.0-beta

    * Removed configuration options that were dedicated to enable the guard
      for selected worlds.
    * The *new* guard simply monitors all worlds selected with the **-W**
      or **-w** argument.

3.2.2-beta

    * removed configuration options
    * added   port check again
    * added   different output formats
"""


# Modules
# ------------------------------------------------

# std
import os
import re
import sys
import time
import socket
import logging
import json

# third party
import termcolor

# local
import emsm
from emsm.core.base_plugin import BasePlugin


# Backward compatibility
# ------------------------------------------------

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = OSError


# Data
# ------------------------------------------------

PLUGIN = "Guard"

log = logging.getLogger(__file__)


# Exceptions
# ------------------------------------------------

class TestFailure(Exception):
    """
    Raised by a guard world test method, if a world did not pass the test.
    """

    def __init__(self, world, test_name, message=None):
        self.world = world
        self.test_name = test_name
        self.message = message
        return None

    def __str__(self):
        if not self.message is None:
            tmp = "The world '{}' did not pass the guard test '{}': {}"\
                  .format(self.world.name(), self.test_name, self.message)
        else:
            tmp = "The world '{}' did not pass the guard test '{}."\
                  .format(self.world.name(), self.test_name)
        return tmp


# Functions
# ------------------------------------------------

def port_is_open(adr, timeout=1, attempts=1):
    """
    Returns `true` if the tcp address *ip*:*port* is reachable.

    Parameters:
        * adr
            The network address tuple of the target.
        * timeout
            Time in seconds waited until a connection attempt is
            considered to be failed.
        * attempts
            Number of port checks done until the *adr* is considered
            to be unreachable.
    """
    for i in range(attempts):
        s = socket.socket()
        s.settimeout(timeout)
        try:
            s.connect(adr)
        except Exception as err:
            pass
        else:
            return True
        finally:
            s.close()
    return False


# Classes
# ------------------------------------------------

class Guard(BasePlugin):

    VERSION = "4.0.0-beta"

    DESCRIPTION = __doc__

    def __init__(self, app, name):
        """
        """
        BasePlugin.__init__(self, app, name)
        self._setup_argparser()

        # We save information about the world health status and occured errors
        # in the guard database. Read more below.
        self._guard_db = None
        self._load_guard_db()
        return None

    def _setup_argparser(self):
        """
        Sets the argument parser up.
        """
        parser = self.argparser()
        parser.description = "Monitors the worlds and reacts on issues."

        # Test selection
        tests_group = parser.add_argument_group(
            title = "tests",
            description = ("You can define which tests are performed, by "
                           "passing one of these arguments. If no test is "
                           "selected, all tests are performed."
                           )
            )
        tests_group.add_argument(
            "--test-status",
            action = "count",
            dest = "guard_test_status",
            help = "Check if the world is online."
            )
        tests_group.add_argument(
            "--test-log",
            action = "count",
            dest = "guard_test_log",
            help = "Check if the logs contain an error."
            )
        tests_group.add_argument(
            "--test-port",
            action = "count",
            dest = "guard_test_port",
            help = "Check if the world's server is reachable."
            )

        # Error action
        parser.add_argument(
            "--error-action",
            action = "store",
            choices = ("none", "stop", "restart"),
            default = "none",
            dest = "guard_error_action",
            help = "Defines the reaction on detected errors."
            )

        # Output
        output_group = parser.add_argument_group(
            title = "output",
            description = ("You can change the output format or restrict the "
                           "amount of printed warnings.")
            )
        output_group.add_argument(
            "--output-format",
            action = "store",
            choices = ("console", "text"),
            default = "console",
            dest = "guard_output_format",
            help = "Defines the output format."
            )
        output_group.add_argument(
            "--output-only-new-warnings",
            action = "count",
            dest = "guard_output_only_new_warnings",
            help = "Prints only new warnings."
            )
        return None

    # Guard database

    """
    We store some data about the worlds health, so that we are able to print a
    warning for the same error only once.

    The guard db is simply a json serialized dictionary, which stores only
    information about worlds which are *already* in trouble:

        {'myworld': {'failed_test': 'status',
                     'test_message': 'world is offline',
                     'time': 1418996881.327088,
                     'warning_printed': False
                     }
         'world2': ...
        }

    A world is removed from this database, as soon as it is restarted or if
    it passes all tests.
    """

    def _guard_db_path(self):
        """
        """
        return os.path.join(self.data_dir(), "errors.json")

    def _load_guard_db(self):
        """
        """
        try:
            with open(self._guard_db_path()) as file:
                self._guard_db = json.load(file)
        except (IOError, FileNotFoundError):
            self._guard_db = dict()
        return None

    def _save_guard_db(self):
        """
        """
        with open(self._guard_db_path(), "w") as file:
            json.dump(self._guard_db, file)
        return None

    # World health checks

    """
    When a test method detects a problem with the world, it raises a
    ``TestFailedError`` exception.
    """

    def _test_status(self, world):
        """
        This test fails, if the world is offline or has been launched multiple
        times.
        """
        pids = world.pids()

        # Check if the world is offline.
        if len(pids) == 0:
            raise TestFailure(world, "status", "world is offline")

        # Check if the world has been launched multiple times.
        if len(pids) >= 2:
            raise TestFailure(
                world, "status", "world has been launched more than one time."
                )
        return None

    def _test_log(self, world):
        """
        This test failes, if the log contains a severe error.
        """
        # Search the log for error records.
        error_re = world.server().log_error_re()
        log = world.latest_log()

        res = re.search(error_re, log)
        if res:
            raise TestFailure(world, "log", str(res))
        return None

    def _test_port(self, world):
        """
        This test fails, if the world's port is not open.
        """
        ip, port = world.address()

        # Check if the EMSM could retrieve the world's address.
        if port is None:
            log.warning("port test for '{}' could not be performed, since the "
                        "world's address could not be retrieved."
                        )
        # Check if we can create a connection to the address and raise an
        # error if not.
        elif not port_is_open((ip, port), timeout=1, attempts=5):
            raise TestFailure(world, "port")
        return None

    def _test(self, world, args):
        """
        Performs the via *args* selected tests on the world, to check if
        everything works fine.
        """
        # Todo: If the number of tests gets bigger, it would be wise
        #       to replace this with loops.

        do_all_tests = not (args.guard_test_status \
                            or args.guard_test_log \
                            or args.guard_test_port
                            )

        if do_all_tests or args.guard_test_status:
            self._test_status(world)
        if do_all_tests or args.guard_test_log:
            self._test_log(world)
        if do_all_tests or args.guard_test_port:
            self._test_port(world)
        return None

    # Error reaction

    def _handle_error(self, world, args):
        """
        The *world* is in trouble. This method reacts on the world's issues
        with the defined *errro_action* in *args*.
        """
        # I assume, that no error happens here.
        if args.guard_error_action == "none":
            pass
        elif args.guard_error_action == "stop":
            world.stop(force_stop=True)
        elif args.guard_error_action == "restart":
            world.restart(force_restart=True)
        return None

    def _guard(self, world, args):
        """
        Tests if the world is in trouble. Reacts if necessairy on issues and
        saves information about the world's status in the guard database.
        """
        try:
            self._test(world, args)
        except TestFailure as err:
            log.warning(err)

            # Note, that *db_record* is actually a reference and not only
            # a copy.
            db_record = self._guard_db.get(world.name(), dict())
            self._guard_db[world.name()] = db_record

            # Handle the error, if we did not react earlier on it
            # or if the error_action changed.
            if db_record.get("error_action") is None \
               or db_record.get("error_action") != args.guard_error_action:
                self._handle_error(world, args)

            # Update the guard database.
            db_record["failed_test"] = err.test_name
            db_record["test_message"] = err.message
            db_record["test_time"] = time.time()
            db_record["error_action"] = args.guard_error_action
            db_record["warning_printed"] = db_record.get("warning_printed", False)
        else:
            # The world is running fine. So we can remove it from the
            # error database, if it was registered.
            if world.name() in self._guard_db:
                self._guard_db.pop(world.name())
        return None

    # Output

    def _print_status_text(self, world, db_record):
        """
        """
        print(world.name())
        print("="*len(world.name()))
        print()
        print("failed_test:  ", db_record["failed_test"])
        print("test_message: ", db_record["test_message"])
        print("test_time:    ", time.ctime(db_record["test_time"]))
        print("error_action: ", db_record["error_action"])
        print()
        print()
        return None

    def _print_status_console(self, world, db_record):
        """
        """
        print(termcolor.colored("{}:".format(world.name()), "cyan"))
        print("\t", "failed_test:  ", db_record["failed_test"])
        print("\t", "test_message: ", db_record["test_message"])
        print("\t", "test_time:    ", time.ctime(db_record["test_time"]))
        print("\t", "error_action: ", db_record["error_action"])
        return None

    def _print_status(self, world, args):
        """
        Prints a warning if the world is in trouble or nothing, if not.
        """
        # Get the status report without altering the database.
        db_record = self._guard_db.get(world.name())

        # Break, if the world is not in trouble.
        # (This means that no records exists.)
        if db_record is None:
            return None

        # Break, if we should not print the same warning twice.
        if db_record.get("warning_printed", False) \
           and args.guard_output_only_new_warnings:
            return None

        # Print the status.
        if args.guard_output_format == "console":
            self._print_status_console(world, db_record)
        elif args.guard_output_format == "text":
            self._print_status_text(world, db_record)

        # Mark the status as already printed.
        db_record["warning_printed"] = True
        return None

    # --

    def run(self, args):
        """
        """
        # Run the guard for all selected worlds in alphabetical order.
        worlds = self.app().worlds().get_selected()
        worlds.sort(key = lambda w: w.name())
        for world in worlds:
            self._guard(world, args)
            self._print_status(world, args)

        # Save any changes made during the run.
        self._save_guard_db()
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
About
-----

This plugin provides a user interface for the server wrapper. It can handle
the server files and their configuration parameters easily.

Download
--------

You can find the latest version of this plugin in the **EMSM**
`GitHub repository <https://github.com/benediktschmitt/emsm>`_.

Configuration
-------------

.. code-block:: ini

    [server]
    update_message = The server is going down for an update.
        Come back soon.

**update_message**

    Message sent to a world before stopping the world due to an server
    update.

Arguments
---------

.. note::

    Make sure to select the server via ``-s, --server``.

.. option:: --usage

    Prints the names of the worlds, powered by a server.

.. option:: --list

    Prints the names of all server supported by the EMSM.

.. option:: --update

    Updates the server software.
"""


# Modules
# ------------------------------------------------

# std
import os
import sys
import logging

# third party
import termcolor

# local
import emsm
from emsm.core.base_plugin import BasePlugin


# Data
# ------------------------------------------------

PLUGIN = "Server"

log = logging.getLogger(__file__)


# Classes
# ------------------------------------------------

class Server(BasePlugin):

    VERSION = "4.0.0-beta"

    DESCRIPTION = __doc__

    def __init__(self, application, name):
        """
        """
        BasePlugin.__init__(self, application, name)

        self._setup_conf()
        self._setup_argparser()
        return None

    def _setup_conf(self):
        """
        Reads all configuration options.
        """
        conf = self.conf()

        self._update_message = conf.get(
            "update_message",
            "The server is going down for an update.\nCome back soon."
            )
        conf["update_message"] = self._update_message
        return None

    def _setup_argparser(self):
        """
        Sets the argument parser of this plugin up.
        """
        parser = self.argparser()

        parser.description = "Manage your server executables"

        # We allow only one argument (keep it simple).
        me_group = parser.add_mutually_exclusive_group()
        me_group.add_argument(
            "--usage",
            action = "count",
            dest = "server_usage",
            help = "Prints all worlds powered by a server."
            )
        me_group.add_argument(
            "--list",
            action = "count",
            dest = "server_list",
            help = "Prints the names of all server supported by the EMSM."
            )
        me_group.add_argument(
            "--update",
            action = "count",
            dest = "server_update",
            help = "Updates the server software."
            )
        return None

    def run(self, args):
        """
        """
        if args.server_list:
            self._print_list()

        else:
            # Sort the server by their names, before running.
            sel_server = self.app().server().get_selected()
            sel_server.sort(key = lambda s: s.name())

            for server in sel_server:
                if args.server_usage:
                    self._print_usage(server)
                elif args.server_update:
                    self._update_server(server)
        return None

    def _print_usage(self, server):
        """
        Prints all worlds that are powered by the server *server*.
        """
        # Get all worlds powered by this server and sort them.
        worlds = self.app().worlds().get_by_pred(
            lambda w: w.server() is server
            )
        online_worlds = list(filter(lambda w: w.is_online(), worlds))
        online_worlds.sort(key = lambda w: w.name())

        offline_worlds = list(filter(lambda w: w.is_offline(), worlds))
        offline_worlds.sort(key = lambda w: w.name())

        # Print the worlds grouped by their current status (offline/online).
        print(termcolor.colored("{}:".format(server.name()), "cyan"))
        print("\t", "* {} worlds".format(len(worlds)))

        print("\t", "* {} online worlds".format(len(online_worlds)))
        if online_worlds:
            for world in online_worlds:
                print("\t\t", "- {}".format(world.name()))

        print("\t", "* {} offline worlds".format(len(offline_worlds)))
        if offline_worlds:
            for world in offline_worlds:
                print("\t\t", "- {}".format(world.name()))
        return None

    def _print_list(self):
        """
        Prints a list with the names of all available server software.
        """
        names = self.app().server().get_names()
        names.sort()

        for name in names:
            print("* {}".format(name))
        return None

    def _update_server(self, server):
        """
        Updates the server *server*.

        All worlds which are currently online and powered by the *server* will
        be stopped and restarted after the update.
        """
        print(termcolor.colored("{}:".format(server.name()), "cyan"))

        # Get all worlds, that are currently running the server.
        worlds = self.app().worlds().get_by_pred(
            lambda w: w.server() is server and w.is_online()
            )
        worlds.sort(key = lambda w: w.name())

        # Stop those worlds.
        try:
            for world in worlds:
                print("\t", "stopping the world '{}' ...".format(world.name()))
                world.stop(message=self._update_message)

        # Do not continue if a world could not be stopped.
        except emsm.core.worlds.WorldStopFailed as err:
            print("\t", termcolor.colored("error:", "red"),
                  "the world '{}' could not be stopped.".format(err.world.name())
                  )
            log.exception(err)

        # Continue with the server update if all worlds are offline.
        # Note, that a ServerIsOnlineError can not occur since we stopped
        # all worlds. (If a world would still be online, we would not have
        # reached this line of code.)
        else:
            print("\t", "reinstalling the server ...")
            try:
                server.reinstall()
            except emsm.core.server.ServerInstallationFailure as err:
                print("\t", termcolor.colored("error:", "red"), err)
                log.exception(err)

        # Restart the worlds.
        finally:
            for world in worlds:
                print("\t", "restarting the world '{}' ..."\
                      .format(world.name()))
                try:
                    world.start()
                except emsm.core.worlds.WorldStartFailed as err:
                    print("\t", termcolor.colored("error:", "red"),
                          "the world '{}' could not be restarted."\
                          .format(err.world.name())
                          )
                    log.exception(err)
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
About
-----

This is a package manager for EMSM plugins. Uninstall and install plugins with
this plugin.

This plugin works only with valid packages and plugins that store its data in
the dedicated paths.

Download
--------

You can find the latest version of this plugin in the EMSM
`GitHub repository <https://github.com/benediktschmitt/emsm>`_.

Arguments
---------

.. option:: --install ARCHIVE

    Installs an new plugin from the archive. If a plugin with the same name
    already exists, the installation will fail.

.. option:: --remove PLUGIN

    Removes the plugin from the EMSM. Please make sure, that no other plugin
    depends on this one.

.. option:: --list

    Lists all loaded plugins.


Package structure
-----------------

The archive that contains the plugin should have the following structure:

.. code-block:: none

    |- foo.tar.bz2
       |- plugin.py
       |- data
          |- bar.txt
          |- bar.csv
          |- ...

During the installation, the path names will be changed to:

.. code-block:: none

    |- EMSM_ROOT
      |- plugins
         |- foo.py    <= plugin.py
      |-plugins_data
         |- foo       <= data
            |- bar.txt
            |- bar.csv
            |- ...

Builder
-------

This plugin comes with an EMSM independent building script for new plugins.
This means, that you can call this script without having the EMSM environment.

Arguments
^^^^^^^^^

.. option:: --create TARGET
.. option:: --source FILE
.. option:: --data DIRECTORY
.. option:: --help, -h

Example
^^^^^^^

Build the plugin *foo*, that comes with a data directory:

.. code-block:: bash

    $ plugin.py --create build/foo --source dev/foo.py --data dev/foo_data
    $ ls build
    ... foo.tar.bz2 ...
"""


# Modules
# --------------------------------------------------

# std
import os
import shutil
import tempfile
import logging
import argparse

# third party
import termcolor

# local
try:
    import emsm
    from emsm.core.base_plugin import BasePlugin
    emsm_context = True
except ImportError:
    BasePlugin = object
    emsm_context = False


# Backward compatibility
# ------------------------------------------------

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = OSError


# Data
# --------------------------------------------------

PLUGIN = "Plugins"

log = logging.getLogger(__file__)


# Exceptions
# --------------------------------------------------

class PluginManagerError(Exception):
    """
    Base class for all other Exceptions in this module.
    """


class PluginInstallError(PluginManagerError):
    """
    Raised if an error occures during the installation.
    """


class PluginAlreadyInstalledError(PluginInstallError):
    """
    Raised if a plugin should be installed that would
    overwrite another plugin that is already installed.
    """

class PluginPackageCorruptedError(PluginInstallError):
    """
    Raised if a plugin package could not be installed
    because it has an invalid structure.
    """


# Data
# --------------------------------------------------

class PluginInstaller(object):
    """
    Installs a new plugin to the EMSM.
    """

    def __init__(self, app, archive_path):
        """
        """
        self._app = app

        # The path to the plugin archive.
        self._archive_path = archive_path

        # The name of the plugin.
        name = os.path.basename(archive_path)
        name = name[:name.find(".")]
        self._name = name

        # Methods to undo the installation progress. If a installation
        # part is complete, the method to undo this "part" is appended
        # to this list, so that we can call the *undo* functions in the
        # correct order.
        self._undo_func = list()
        return None

    def archive_path(self):
        """
        Returns the path to the archive that contains the plugin that should be
        installed.
        """
        return self._archive_path

    def plugin_name(self):
        """
        Returns the *name* of the plugin.
        """
        return self._name

    def _install_plugin_module(self, package_path):
        """
        Installs the *plugin.py* module in *package_path* into the main
        plugin directory of the EMSM and renames it.

        See also:
            * name()
        """
        package_module_path = os.path.join(package_path, "plugin.py")

        # Check if the plugin module exists.
        if not os.path.exists(package_module_path):
            raise PluginPackageCorruptedError(
                "the package does not contain a *plugin.py* file."
                )

        # Check if it is a file.
        if not os.path.isfile(package_module_path):
            raise PluginPackageCorruptedError(
                "the *plugin.py* file in the package is NOT a file."
                )

        # Copy the plugin.py file to the EMSM *plugins* directory.
        emsm_module_path = os.path.join(
            self._app.paths().plugins_dir(), self._name + ".py"
            )
        shutil.move(package_module_path, emsm_module_path)

        # Add the undo function.
        self._undo_func.append(self._undo_plugin_module)
        return None

    def _undo_plugin_module(self):
        """
        Removes the plugin module previously installed with
        ``_install_plugin_module()``.
        """
        emsm_module_path = os.path.join(
            self._app.paths().plugins_dir(), self._name + ".py"
            )
        try:
            os.remove(emsm_module_path)
        except OSError:
            # We can't do anything ...
            pass
        return None

    def _install_plugin_data(self, package_path):
        """
        Copies the *data* directory from *package_path* to the EMSM
        plugins_data directory and renames it.

        See also:
            * name()
        """
        package_data_dir = os.path.join(package_path, "data")
        emsm_data_dir = self._app.paths().plugin_data_dir(self._name)

        # Check if the package contains a data directory.
        if not os.path.exists(package_data_dir):
            # Break, since the data directory is not required..
            return None

        # Check if the data directory is a *directory*.
        if not os.path.isdir(package_data_dir):
            raise PluginPackageCorruptedError(
                "the *data* directory in the package is NOT a directory."
                )

        # Make sure there is no directory for this plugin in EMSM_ROOT/plugins_data.
        if os.path.exists(emsm_data_dir):
            raise PluginInstallError(
                "the EMSM contains already a plugin directory for '{}'."\
                .format(self._name)
                )

        # Move the data directory to the EMSM *plugins_data* directory.
        shutil.move(package_data_dir, emsm_data_dir)

        # Add the undo function.
        self._undo_func.append(self._undo_plugin_data)
        return None

    def _undo_plugin_data(self):
        """
        Removes the plugin data directory previously installed with
        ``_install_plugin_data_dir()``.
        """
        try:
            shutil.rmtree(self._app.paths().plugin_data_dir(self._name))
        except OSError:
            # We can't do anything ...
            pass
        return None

    def _undo(self):
        """
        Undos all installation steps done till now.
        """
        for func in reversed(self._undo_func):
            func()
        return None

    def _install(self):
        """
        The raw installation progress, that might throw some exceptions.

        Exceptions:
            * PluginInstallError
            * PluginPackageCorruptedError
            * PluginAlreadyInstalledError
        """
        self._undo_func = list()

        # Check if there is already a plugin with that name.
        if self._app.plugins().plugin_is_available(self._name):
            raise PluginAlreadyInstalledError(
                "the plugin '{}' is already installed.".format(self._name)
                )

        # Unpack the archive in a temporary folder.
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                shutil.unpack_archive(
                    filename = self._archive_path,
                    extract_dir = tmp_dir
                    )
            except (ValueError, FileNotFoundError, shutil.ReadError) as err:
                raise PluginInstallError(
                    "the plugin archive '{}' could not be uncompressed."\
                    .format(self._name)
                    )
            else:
                try:
                    self._install_plugin_module(tmp_dir)
                    self._install_plugin_data(tmp_dir)
                except:
                    self._undo()
                    raise
        return None

    def install(self):
        """
        Like *_install()*, but catches the exceptions, prints the progress and
        creates more log entries.
        """
        print("package_path: {}".format(self._archive_path))
        print("package_name: {}".format(self._name))
        print("installing package ...")

        log.info("installing '{}' from '{}' ..."\
                 .format(self._name, self._archive_path)
                 )

        try:
            self._install()
        except PluginAlreadyInstalledError as err:
            print(termcolor.colored("error:", "red"),
                  "a plugin with the name '{}' is already installed."\
                  .format(self._name)
                  )
            log.error(err)
        except PluginPackageCorruptedError as err:
            print(termcolor.colored("error:", "red"),
                  "the package is not a valid plugin package."
                  )
            log.error(err)
        except PluginInstallError as err:
            print(termcolor.colored("error:", "red"), err)
            log.error(err)
        else:
            print("installation complete.")
            log.info("installation of '{}' complete.".format(self._name))
        return None


class Plugins(BasePlugin):

    VERSION = "4.0.0-beta"

    DESCRIPTION = __doc__

    def __init__(self, app, name):
        """
        """
        BasePlugin.__init__(self, app, name)

        self._setup_argparser()
        return None

    def _setup_argparser(self):
        """
        Sets the argument parser up.
        """
        parser = self.argparser()
        parser.description = "Install and remove EMSM plugins"

        # We want to allow only one action per run.
        me_group = parser.add_mutually_exclusive_group()

        me_group.add_argument(
            "--install",
            action = "store",
            dest = "plugins_install",
            metavar = "ARCHIVE",
            help = "Install a new plugin from the given archive."
            )

        # Get the names of all loaded plugins, excluding THIS plugin
        # (the plugins.py plugin).
        plugin_names = [name \
                        for name in self.app().plugins().get_plugin_names()
                        if name != self.name()]
        me_group.add_argument(
            "--uninstall",
            action = "store",
            dest = "plugins_uninstall",
            metavar = "PLUGIN",
            choices = plugin_names,
            help = "Uninstall the plugin."
            )

        me_group.add_argument(
            "--list",
            action = "count",
            dest = "plugins_list",
            help = "Lists all loaded plugins."
            )
        return None

    def run(self, args):
        """
        """
        if args.plugins_install:
            installer = PluginInstaller(self.app(), args.plugins_install)
            installer.install()

        elif args.plugins_uninstall:
            plugin = self.app().plugins().get_plugin(args.plugins_uninstall)
            plugin.uninstall()

        elif args.plugins_list:
            self._list_plugins()
        return None

    def _list_plugins(self):
        """
        Lists all loaded plugins.
        """
        # We list the plugins in alphabetical order.
        plugins = self.app().plugins().get_plugin_names()
        plugins.sort()

        if not plugins:
            print("- no plugins loaded -")
        else:
            for name in plugins:
                print("* {}".format(name))
        return None


# Main
# ------------------------------------------------

if __name__ == "__main__":

    # Parse the command line arguments.
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--create",
        required = True,
        help = "The target location of the packed archive."
        )
    parser.add_argument(
        "--source",
        required = True,
        help = "The source file of the plugin."
        )
    parser.add_argument(
        "--data",
        help = "The data directory, that should be included."
        )
    args = parser.parse_args()

    # Create the archive
    with tempfile.TemporaryDirectory() as temp_dir:

        # Copy the plugin module and the data directory to the temp_dir.
        print("copying plugin module ...")
        shutil.copy(args.source, os.path.join(temp_dir, "plugin.py"))

        if args.data:
            print("copying plugin data dir ...")
            shutil.copytree(args.data, os.path.join(temp_dir, "data"))

        # Create the archive by zipping the temp_dir.
        print("compressing archive ...")
        archive = shutil.make_archive(
            base_name = args.create,
            format = "bztar",
            root_dir = temp_dir,
            base_dir = "./"
            )
        print("created '{}'.".format(archive))



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# ------------------------------------------------

# std
import os

# emsm
import emsm


# Backward compatibility
# ------------------------------------------------

try:
    FileExistsError
except NameError:
    FileExistsError = OSError


# Data
# ------------------------------------------------

__all__ = [
    "Pathsystem"
    ]


# Classes
# ------------------------------------------------

class Pathsystem(object):
    """
    Manages the paths to the different files and directories of the
    application.

    The EMSM distinguishes two primary folders:
    The *instance* folder, where the worlds, server, configuration and plugins
    of the user are placed. The *instance* folder can actually be considered to
    be the working directory of the EMSM.
    On the other side is the EMSM installation directory. The *emsm* directory.
    This directory is usually placed in Python's third party library folder
    (site-packages) and contains the EMSM core application and the core plugins.

    #. *emsm* directory:

        .. code-block:: none

            |- emsm
                |- core
                    |- lib
                        |- ...
                    |- __init__.py
                    |- application.py
                    |- argparse_.py
                    |- base_plugin.py
                    |- ...
                |- plugins
                    |- __init__.py
                    |- backups.py
                    |- emsm.py
                    |- guard.py
                    |- ...
                |- __init__.py

    #. *instance* folder:

        .. code-block:: none

            |- instance             # usually /opt/minecraft
                |- conf             # the emsm configuration files
                    |- main.conf
                    |- server.conf
                    |- worlds.conf
                |- plugins          # the user's plugins.
                    |- __init__.py
                    |- myawesomeplugin.py
                    |- ...
                |- plugins_data     # data generated or needed by the plugins
                    |- backups
                    |- emsm
                    |- guard
                    |- ...
                    |- myawesomeplugin
                    |- ...
                |- server           # the server executables
                    |- minecraft_server.jar
                    |- craftbukkit.jar
                    |- ...
                |- worlds           # the data of the worlds (minecraft map, ...)
                    |- foo
                        |- server.properties
                        |- ...
                    |- bar
                        |- server.properties
                        |- ...
                |- logs
                    |- emsm.log
                    |- emsm.log.1
                    |- ...
                |- minecraft.py
    """

    def __init__(self, instance_dir):
        """
        """
        self._instance_dir = instance_dir
        self._emsm_dir = os.path.dirname(emsm.__file__)
        return None

    def create(self):
        """
        Creates the folders used by the EMSM Application.

        This method should only be called, after the EMSM downgraded it's
        priviliges.
        """
        def make_dir(path):
            """
            Creates the directory *path* if it does not exist or
            returns silently, if it does.
            """
            try:
                os.makedirs(path, exist_ok=True)
            except OSError:
                pass
            return None

        make_dir(self.instance())
        make_dir(self.conf())
        make_dir(self.plugins())
        make_dir(self.plugins_data())
        make_dir(self.server())
        make_dir(self.worlds())
        make_dir(self.logs())
        return None

    # EMSM

    def emsm(self):
        """
        Returns the path to the *emsm* installation directory. This folder
        is usually located in Python's ``site-packages`` directory.
        """
        return self._emsm_dir

    def emsm_core(self):
        """
        Returns the path to the *emsm.core* directory.
        """
        return os.path.join(self._emsm_dir, "core")

    def emsm_plugins(self):
        """
        Returns the path to the *emsm.plugins* directory. The directory contains
        the core plugins like *worlds*, *server*, *backups*, ...
        """
        return os.path.join(self._emsm_dir, "plugins")

    # Instance

    def instance(self):
        """
        The *instance* folder contains all data generated by the EMSM
        application and the minecraft worlds.
        """
        return self._instance_dir

    def conf(self):
        """
        Contains the configuration files of the EMSM. **Not** the configuration
        for the minecraft worlds. These are still located in the world folder.

        The directory contains the *main.conf*, *server.conf* and
        *worlds.conf* file.

        The directory is located in the *instance* folder.
        """
        return os.path.join(self._instance_dir, "conf")

    def plugins(self):
        """
        Contains all user plugins and plugin packages.

        The directory is located in the *instance* folder.

        .. seealso:: :meth:`emsm_plugins`
        """
        return os.path.join(self._instance_dir, "plugins")

    def plugins_data(self):
        """
        The directory that contains the data generated or used by all plugins.

        The directory is located in the *instance* folder.

        .. seealso:: :meth:`plugin_data``
        """
        return os.path.join(self._instance_dir, "plugins_data")

    def plugin_data(self, plugin_name):
        """
        This directory contains all data of the plugin with the name
        *plugin_name*.

        The directory is a subfolder of :meth:`plugins_data`.
        """
        return os.path.join(self.plugins_data(), plugin_name)

    def server(self):
        """
        This directory contains all server executables specified in
        the ``server.conf`` configuration file.

        The directory is located in the *instance* folder.

        .. seealso:: :meth:`server_`
        """
        return os.path.join(self._instance_dir, "server")

    def server_(self, server_name):
        """
        This directory contains the server software for the server with the
        name *server_name*.

        The directory is located in the :meth:`server` folder.

        .. todo::

            * Try to find better names for *server* and *server_*. They are hard
              to distinguish.
        """
        return os.path.join(self.server(), server_name)

    def worlds(self):
        """
        Contains for each world in ``worlds.conf`` one folder that contains
        the data generated by the minecraft server.

        The directory is located in the *instance* folder.

        .. seealso:: :meth:`world`
        """
        return os.path.join(self._instance_dir, "worlds")

    def world(self, world_name):
        """
        This directory contains the data generated by the minecraft server
        which powers the world *world_name*.
        It contains among others those files:

            * :file:`server.properties`
            * :file:`ops.json`
            * :file:`whitelist.json`

        Furthermore, it is a child of :meth:`worlds`.
        """
        return os.path.join(self.worlds(), world_name)

    def logs(self):
        """
        Contains the EMSM log files.

        Note, that this is NOT the log directory of the minecraft server.
        """
        return os.path.join(self._instance_dir, "log")



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# ------------------------------------------------

# std
import logging
import logging.handlers
import os
import queue


# ...
# ------------------------------------------------

__all__ = [
    "Logger"
    ]


# Classes
# ------------------------------------------------
class Logger(object):
    """
    Sets the *root* :class:`logging.Logger` up.

    The EMSM logger queues all log records until the :file:`emsm.log` can be
    acquired without side effects. This is the case, when the
    :class:`~emsm.core.application.Application` acquired the *file lock*.
    The queued records are then pushed to the :file:`emsm.log`.

    The EMSM logging stategy requires, that each module uses its own
    logger instance:

    .. code-block:: python

        >>> import logging
        >>> log = logging.getLogger(__file__)
    """

    def __init__(self, app):
        """
        """
        self._app = app
        self._log_dir = app.paths().logs()

        # The root logger
        self._root_log = logging.getLogger()
        self._root_log.setLevel(logging.INFO)

        # The EMSM log format.
        self._log_format = logging.Formatter(
            fmt="[{asctime}][{levelname:8}][{module:25}] {message}",
            datefmt="%Y-%m-%d %H:%M:%S", style="{"
            )

        # Used to store the log records until we can access the *emsm.log*.
        # As soon as *emsm.log* is opened, we write all records in the queue
        # to the log file.
        # These attributes are set to None, as soon as the log file is opened.
        self._log_queue = queue.Queue()
        self._log_queue_handler = logging.handlers.QueueHandler(self._log_queue)
        self._root_log.addHandler(self._log_queue_handler)

        # The FileHandler for the EMSM log file (usually *emsm.log*)
        # The file is opened, as soon as the Application acquired the file lock.
        #
        # See also:
        #   * setup()
        self._log_file_handler = None
        return None

    def setup(self):
        """
        Opens the :file:`emsm.log` and pushes all queued log recods to the log
        file.

        .. hint::

            This method requires that the Application aquired the file lock.
        """
        # We use the rotating file handler so that the logfiles are
        # automatically compressed, when they are bigger than 10mb.
        self._log_file_handler = logging.handlers.RotatingFileHandler(
            os.path.join(self._log_dir, "emsm.log"),
            maxBytes = 10*1024**2,
            backupCount = 5
            )
        self._log_file_handler.setFormatter(self._log_format)

        # Push the queued records to the file handler.
        while not self._log_queue.empty():
            record = self._log_queue.get(block=False)
            self._log_file_handler.handle(record)

        # Add the file handler to the root logger and remove the queue handler.
        self._root_log.addHandler(self._log_file_handler)

        self._root_log.removeHandler(self._log_queue_handler)
        self._log_queue_handler = None
        self._log_queue = None
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# -----------------------------------------------

# std
from setuptools import setup

# local
try:
    import emsm
except ImportError:
    emsm = None


# Setup
# -----------------------------------------------

long_description = open("README.rst").read()

requirements = [
    line.strip()
    for line in open("emsm/requirements.txt")
    if line.strip()
    ]

version = emsm.core.VERSION if emsm else "- n/a -"

setup(
    name = "emsm",
    version = version,
    url = "https://github.com/benediktschmitt/emsm",
    license = "MIT License",
    author = "Benedikt Schmitt",
    author_email = "benedikt@benediktschmitt.de",
    description = "A lightweight, easy to extend mineraft server manager",
    long_description = long_description,
    packages = ["emsm", "emsm.core", "emsm.core.lib", "emsm.plugins"],
    include_package_data = True,
    platforms = "LINUX",
    install_requires = requirements,
    classifiers = [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.2",
        "Topic :: Games/Entertainment",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities"
    ]
)



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
About
-----

Works as interface between the linux *initd service* and the EMSM.

Download
--------

You can find the latest version of this plugin in the EMSM
`GitHub repository <https://github.com/benediktschmitt/emsm>`_.

Installation
------------

You only have to create the ``init.d`` script :file:`/etc/init.d/minecraft`:

..  literalinclude:: ../../../emsm/core/initd_script.sh
    :language: bash
    :linenos:

.. code-block:: bash

    $ sudo chmod +x /etc/init.d/minecraft
    $ sudo update-rc.d minecraft defaults
    $ sudo update-rc.d minecraft enable

Configuration
-------------

worlds.conf
^^^^^^^^^^^

.. code-block:: ini

    [DEFAULT]
    enable_initd = yes

    [foo]
    enable_initd = no

**enable_initd**

    If ``yes``, the autostart/-stop is enabled.

If you want to enable *init.d* for all worlds, use the *DEFAULT* section.

Arguments
---------

.. option:: --start

    Starts all worlds, where the *enable_initd* configuration value is true.

.. option:: --stop

    Stops all worlds, where the *enable_initd* configuration value is true.
    Note, that this will always **force** the stop of the world, since the
    process is killed anyway during system shutdown.

.. option:: --restart

    Forces the restart of all worlds which has *enable_initd* enabled.

Exit code
---------

The exit code is set to:

* 0 if no error occured.
* 2 if an error occured.
"""


# Modules
# ------------------------------------------------

# std
import logging

# third party
import blinker
import termcolor

# local
import emsm
from emsm.core.base_plugin import BasePlugin


# Data
# ------------------------------------------------

PLUGIN = "InitD"

log = logging.getLogger(__file__)


# Classes
# ------------------------------------------------

class InitD(BasePlugin):

    VERSION = "4.0.0-beta"

    DESCRIPTION = __doc__

    # Emitted when initd is called with the *--start* argument.
    on_initd_start = blinker.signal("initd_start")

    # Emitted when initd is called with the *--stop* argument.
    on_initd_stop = blinker.signal("initd_stop")

    # Emitted when initd is called with the *--restart* argument.
    on_initd_restart = blinker.signal("initd_restart")

    def __init__(self, app, name):
        """
        """
        BasePlugin.__init__(self, app, name)

        self._setup_argparser()
        return None

    def _setup_argparser(self):
        """
        Sets the argument parser up.
        """
        parser = self.argparser()

        parser.description = "InitD interface"

        # Allow only one runlevel to the same time.
        me_group = parser.add_mutually_exclusive_group()
        me_group.add_argument(
            "--start",
            action = "count",
            dest = "initd_start",
            help = "Starts all worlds for which initd support is enabled."
            )
        me_group.add_argument(
            "--stop",
            action = "count",
            dest = "initd_stop",
            help = "Stop all worlds for which initd support is enabled."
            )
        me_group.add_argument(
            "--restart",
            action = "count",
            dest = "initd_restart",
            help = "Restart all worlds for which initd support is enabled."
            )
        me_group.add_argument(
            "--status",
            action = "count",
            dest = "initd_status",
            help = "Prints the status of all initd managed worlds."
            )
        return None

    def _uninstall(self):
        """
        Makes sure the configuration options added to the *world.conf* are
        removed.
        """
        # Clean the worlds.conf up.
        world_conf = self.app().conf().worlds()
        for section in world_conf:
            world_conf.remove_option(section, "enable_initd")
        return None

    def _initd_worlds(self):
        """
        Returns all worlds where *enable_initd* is true.
        The worlds are sorted by their names.

        See also:
            * WorldWrapper.conf()
        """
        worlds = self.app().worlds().get_by_pred(
            lambda w: w.conf().getboolean("enable_initd", False)
            )
        worlds.sort(key = lambda w: w.name())
        return worlds

    def _start(self):
        """
        Starts all worlds if *enable_initd* is true.
        """
        # We create the unformatted messages here to increase readability.
        raw_msg = "[ {status} ] starting the minecraft world '{{world_name}}'"
        pre_msg = raw_msg.format(status="... ")
        fail_msg = raw_msg.format(status=termcolor.colored("fail", "red"))
        ok_msg = raw_msg.format(status=termcolor.colored("ok  ", "green"))

        # Start the worlds.
        log.info("initd start ...")

        for world in self._initd_worlds():
            print(pre_msg.format(world_name=world.name()), end="\r")
            try:
                world.start()
            except emsm.core.worlds.WorldStartFailed as err:
                print(fail_msg.format(world_name=world.name()))
                self.app().set_exit_code(2)
            else:
                print(ok_msg.format(world_name=world.name()))

        log.info("initd start done.")
        return None

    def _stop(self):
        """
        Stops all worlds if *enable_initd* is true.
        """
        # We create the unformatted messages here to increase readability.
        raw_msg = "[ {status} ] stopping the minecraft world '{{world_name}}'"
        pre_msg = raw_msg.format(status="... ")
        fail_msg = raw_msg.format(status=termcolor.colored("fail", "red"))
        ok_msg = raw_msg.format(status=termcolor.colored("ok  ", "green"))

        # Stop the worlds.
        log.info("initd stop ...")

        for world in self._initd_worlds():
            print(pre_msg.format(world_name=world.name()), end="\r")
            try:
                # Because the process is killed anyway, we force it here.
                world.stop(force_stop=True)
            except emsm.core.worlds.WorldStopFailed as err:
                print(fail_msg.format(world_name=world.name()))
                self.app().set_exit_code(2)
            else:
                print(ok_msg.format(world_name=world.name()))

        log.info("initd stop done.")
        return None

    def _restart(self):
        """
        Forces the restart of all worlds which has *enable_initd* enabled.
        """
        # We create the unformatted messages here to increase readability.
        raw_msg = "[ {status} ] restarting the minecraft world '{{world_name}}'"
        pre_msg = raw_msg.format(status="... ")
        fail_msg = raw_msg.format(status=termcolor.colored("fail", "red"))
        ok_msg = raw_msg.format(status=termcolor.colored("ok  ", "green"))

        # Restart the worlds.
        log.info("initd restart ...")

        for world in self._initd_worlds():
            print(pre_msg.format(world_name=world.name()), end="\r")
            try:
                # Because the process is killed anyway, we force it here.
                world.restart(force_restart=True)
            except emsm.core.worlds.WorldStopFailed as err:
                print(fail_msg.format(world_name=world.name()))
                self.app().set_exit_code(2)
            except emsm.core.worlds.WorldStartFailed as err:
                print(fail_msg.format(world_name=world.name()))
                self.app().set_exit_code(2)
            else:
                print(ok_msg.format(world_name=world.name()))

        log.info("initd restart done.")
        return None

    def _status(self):
        """
        Prints the status of all worlds where *enable_initd* is true.
        """
        # We create the unformatted messages here to increase readability.
        fail_msg = "[ {status} ] the minecraft world '{{world_name}}' is offline."\
                   .format(status=termcolor.colored("fail", "red"))
        ok_msg = "[ {status} ] the minecraft world '{{world_name}}' is online."\
                 .format(status=termcolor.colored("ok  ", "green"))

        # Print the status the worlds.
        for world in self._initd_worlds():
            if world.is_online():
                print(ok_msg.format(world_name=world.name()))
            else:
                print(fail_msg.format(world_name=world.name()))
        return None

    def run(self, args):
        """
        """
        if args.initd_start:
            self._start()
            InitD.on_initd_start.send()
        elif args.initd_stop:
            self._stop()
            InitD.on_initd_stop.send()
        elif args.initd_restart:
            self._restart()
            InitD.on_initd_restart.send()
        elif args.initd_status:
            self._status()
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
About
-----

Provides information about the EMSM itself, like the version and simplifies the
EMSM update.

Download
--------

You can find the latest version of this plugin in the EMSM
`GitHub repository <https://github.com/benediktschmitt/emsm>`_.

Arguments
---------

.. option:: --version

    Shows the current EMSM version number.

.. option:: --license

    Shows the EMSM license.

.. option:: --check-update

    Simulates an EMSM update.
"""


# Modules
# ------------------------------------------------

# std
import os
import re
import urllib.request
import logging
import collections

# local
import emsm
from emsm.core.base_plugin import BasePlugin


# Data
# ------------------------------------------------

PLUGIN = "EMSM"

log = logging.getLogger(__file__)


# Classes
# ------------------------------------------------

class Updater(object):
    """
    Manages the EMSM update.
    """

    def __init__(self, app):
        self._app = app

        # Cache some values.
        self._latest_version = None
        return None

    # version number
    # ^^^^^^^^^^^^^^

    def _split_version(self, version):
        """
        Splits the version string in the following way:

            >>> version
            "3.0.1-beta"
            >>> _split_version(version)
            [3, 0, 1, "beta"]
        """
        [major, minor, patch, phase] = re.split("[.-]", version)
        major = int(major)
        minor = int(minor)
        patch = int(patch)
        return dict(major=major, minor=minor, patch=patch, phase=phase)

    def build_version_string(self, version):
        """
        Builds the version string from the version dictionary *version*.
        """
        return "{major}.{minor}.{patch}-{phase}".format(**version)

    def current_version(self):
        """
        Returns the current version of the EMSM.
        """
        return self._split_version(emsm.core.version.VERSION)

    def latest_version(self):
        """
        Returns the latest EMSM version.
        """
        # Check if we already downloaded this value.
        if not self._latest_version is None:
            return self._latest_version

        latest_version_py_url = "https://pypi.python.org/pypi?:action=doap&name=emsm"
        with urllib.request.urlopen(latest_version_py_url) as file:
            data = file.read()

        # Isolate the version node and get its value.
        latest_version = re.findall(b"<revision>(.*?)</revision>", data)
        latest_version = latest_version[0]
        latest_version = latest_version.decode()
        latest_version = self._split_version(latest_version)

        # Save the value to avoid multiple http requests. I assume, that the
        # version number does not change during one EMSM run.
        self._latest_version = latest_version
        return latest_version

    def version_is_outdated(self):
        """
        Returns ``True`` if the current EMSM installation is out of date.
        """
        latest_version = self.latest_version()
        current_version = self.current_version()

        # Check the version *number* before checking the stadium of the
        # software (beta, release, ...)
        if latest_version["major"] > current_version["major"]:
            return True
        elif latest_version["minor"] > current_version["minor"]:
            return True
        elif latest_version["patch"] > current_version["patch"]:
            return True

        # EMSM has only two software development states: *beta* and *release*.
        # A beta version of the EMSM is usually, a release candidate.
        if latest_version["phase"] == "release" \
           and current_version["phase"] == "beta":
            return True
        return False


class EMSM(BasePlugin):

    VERSION = "4.0.0-beta"

    DESCRIPTION = __doc__

    def __init__(self, app, name):
        super().__init__(app, name)

        self._setup_conf()
        self._setup_argparser()
        return None

    def _setup_conf(self):
        """
        """
        # When using the configuration for this plugin, make sure to not
        # interfere with the global emsm configuration section (this section
        # contains the *user* and *timeout* options).
        return None

    def _setup_argparser(self):
        """
        Sets the argument parser up.
        """
        parser = self.argparser()

        parser.description = "A manager for the EMSM itself"

        # We want to allow only one action per run, so we put everything in a
        # mutually exclusive group.
        me_group = parser.add_mutually_exclusive_group()
        me_group.add_argument(
            "--version",
            action = "count",
            dest = "emsm_version",
            help = "Shows the current EMSM version number."
            )
        me_group.add_argument(
            "--license",
            action = "count",
            dest = "emsm_license",
            help = "Shows the EMSM license.",
            )
        me_group.add_argument(
            "--check-update",
            action = "count",
            dest = "emsm_check_update",
            help = "Simulates an EMSM update."
            )
        return None

    def _action_version(self):
        """
        Displays the current emsm version number and hash value of the last git
        commit.
        """
        print(emsm.core.version.VERSION)
        return None

    def _action_license(self):
        """
        Prints the EMSM license.
        """
        print(emsm.core.license_.LICENSE)
        return None

    def _action_check_update(self):
        """
        Checks if the EMSM can be updated and simulates the update.
        """
        updater = Updater(self.app())
        if updater.version_is_outdated():
            print("Your version it outdated.")
            print("Current version:",
                  updater.build_version_string(updater.current_version()))
            print("Latest version: ",
                  updater.build_version_string(updater.latest_version()))
        else:
            print("Your version is up to date.")
        return None

    def run(self, args):
        """
        """
        if args.emsm_version:
            self._action_version()
        elif args.emsm_license:
            self._action_license()
        elif args.emsm_check_update:
            self._action_check_update()
        return None
